from typing import Protocol, Any

from pandas import DataFrame


class DataframeSaver(Protocol):
    @staticmethod
    def save(folder, df: DataFrame, configs: tuple[Any], *args, **kwargs):
        """Save a file with the given name and parameters."""
        pass



from NL2SQLEvaluator.evaluator_nodes.bird_ex_evaluator import BirdEXEvaluator
from NL2SQLEvaluator.evaluator_nodes.ex_with_card import EXEvaluator
from NL2SQLEvaluator.evaluator_nodes.qatch_metrics import QatchMetric



from NL2SQLEvaluator.evaluator_nodes.evaluator_protocol import EvaluateTask
from NL2SQLEvaluator.evaluator_nodes.utils import get_majority_voting_values
from NL2SQLEvaluator.node_registry import register_node


@register_node()
class BirdEXEvaluator:
    """Implementation of the standard EX from BIRD: https://bird-bench.github.io/"""
    def execute_metric(
            self,
            tasks: list[EvaluateTask],
            *args,
            **kwargs
    ) -> list[float]:
        # self._validate_inputs(multiple_tasks_preds, multiple_tasks_tar)
        results = [self._ex(task) for task in tasks]
        return results

    def _ex(self, task: EvaluateTask) -> float:
        target = frozenset(task.target[0])
        majority_vote = get_majority_voting_values(task.predictions, count_cardinality_in_row=False)
        if majority_vote is None:
            return 0.0
        return 1.0 if frozenset(target) == majority_vote else 0.0

    @staticmethod
    def _validate_inputs(multiple_tasks_preds, multiple_tasks_tar) -> None:
        if not (len(multiple_tasks_preds) == len(multiple_tasks_tar)):
            raise ValueError(
                "Lengths must match: target_queries, llm_predictions, db_files."
            )
        for ref in multiple_tasks_tar:
            if not len(ref) == 1:
                raise ValueError(
                    "For Execution Accuracy the target must have len 1. Otherwise use Ambiguous Execution Accuracy."
                )



import pandas as pd

from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import DataInput, enforce_input_type
from NL2SQLEvaluator.node_registry import register_node


@register_node()
class ReadBird:
    @staticmethod
    def read(file_path: str, base_db_path: str, *args, **kwargs) -> list[DataInput]:
        if file_path.endswith(".json"):
            df = pd.read_json(file_path)
        elif file_path.endswith(".csv"):
            df = pd.read_csv(file_path)
        else:
            import datasets
            df = datasets.load_dataset(file_path).to_pandas()

        df['db_file'] = df['db_id'].map(lambda row: f"{base_db_path}/{row}/{row}.sqlite")
        df['target_query'] = df['SQL']
        values = df.to_dict(orient='records')
        enforce_input_type(values)
        return values



import json
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import Any

import pandas as pd
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import register_node

logger = get_logger(__name__)


@register_node()
class JSONSaver:
    @staticmethod
    def save(folder: Path, df: pd.DataFrame, configs: tuple[Any], *args, **kwargs):
        from datetime import datetime
        date_str = datetime.now().strftime("%Y-%m-%d")
        hours = datetime.now().strftime("%H-%M-%S")
        path = Path(folder) / date_str / hours
        path.mkdir(parents=True, exist_ok=True)

        df_path = path / 'df.json'
        df.to_json(df_path, orient='records', indent=4)
        logger.info(f"Saved DF in {df_path} as JSON")

        for config in configs:
            config_name = config.__class__.__name__ if is_dataclass(config) else type(config).__name__
            config_path = path / f"{config_name}.json"

            config_data = asdict(config) if is_dataclass(config) else config
            with open(config_path, "w") as f:
                json.dump(config_data, f, indent=4, default=str)
            logger.info(f"Saved config in {config_path} as JSON")



import hashlib
from typing import Self, Any, Iterator, Protocol

import sqlglot
from pydantic import BaseModel, model_validator


class NotFoundInCacheError(Exception):
    """Raised when an item is not found in the cache."""
    pass


class OutputTable(BaseModel):
    rows: list[tuple | list]

    @model_validator(mode='after')
    def forbid_inner_lists(self) -> Self:
        for i, row in enumerate(self.rows):
            for j, val in enumerate(row):
                if isinstance(val, list | tuple | set | dict):
                    raise TypeError(
                        f"rows[{i}][{j}] is a list, which is forbidden"
                    )
        return self

    def __len__(self):
        return len(self.rows)

    def __call__(self, index: int | slice | None = None) -> Any:
        """
        Call to access rows:
        - no argument -> return the full rows list
        - int -> return the row at that index (supports negative indices)
        - slice -> return a sublist of rows
        """
        if index is None:
            return self.rows
        if isinstance(index, (int, slice)):
            return self.rows[index]
        raise TypeError("index must be an int, slice, or None")

    def __getitem__(self, item):
        return self.rows[item]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.rows)

    def __contains__(self, item: Any) -> bool:
        return item in self.rows

    def compress(self) -> bytes:
        """Compress the result to save space in cache."""
        # Implement compression logic if needed
        import pickle
        return pickle.dumps(self)

    def decompress(self) -> Self:
        import pickle
        return pickle.loads(self.rows[0][0]) if self.rows and self.rows[0] else self


class DataToFetch(BaseModel):
    db_id: str
    query: str
    dialect: str = "sqlite"
    hash_key: Any = None

    @model_validator(mode='after')
    def create_hash_key(self) -> Self:
        try:
            self.query = sqlglot.transpile(self.query, self.dialect, identity=True)[0]
        except Exception:
            ...  # keep original query if transpile fails
        value = f"{self.db_id}|{self.query}"
        self.hash_key = self.hash_key or hashlib.sha256(value.encode("utf-8")).hexdigest()
        return self


class DataToCache(DataToFetch):
    result: OutputTable


class SQLCacheProtocol(Protocol):
    @staticmethod
    def set_in_cache(db_file: str, data_to_cache: list[DataToCache]) -> None:
        """Save a file with the given name and parameters."""
        ...

    @staticmethod
    def get_from_cache(db_file: str, data_to_fetch: list[DataToFetch]) -> list[DataToCache | NotFoundInCacheError]:
        """Retrieve a file with the given name and parameters."""
        ...



from typing import Protocol

from vllm import SamplingParams

from NL2SQLEvaluator.config import ModelArgs
from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import ChatMessageHF


class PredictorProtocol(Protocol):
    def infer(self,
              model_name: str,
              multiple_tasks_messages: list[ChatMessageHF],
              sampling_params: SamplingParams,
              *args,
              **kwargs) -> list[list[str]]:
        ...


def generate_predictions(predictor: PredictorProtocol,
                         model_name: str,
                         multiple_tasks_messages: list[ChatMessageHF],
                         model_args: ModelArgs,
                         *args,
                         **kwargs) -> list[list[str]]:
    sampling_params = create_sampling_params(model_args)
    return predictor.infer(
        model_name=model_name,
        multiple_tasks_messages=multiple_tasks_messages,
        sampling_params=sampling_params,
        *args,
        **kwargs
    )


def create_sampling_params(model_args: ModelArgs) -> SamplingParams:
    return SamplingParams(
        n=model_args.number_of_completions,
        repetition_penalty=model_args.repetition_penalty,
        temperature=model_args.temperature,
        top_p=model_args.top_p,
        top_k=model_args.top_k,
        max_tokens=model_args.max_new_tokens,
    )



from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import NotFoundInCacheError, OutputTable, DataToFetch, \
    DataToCache
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError, ExecuteTask
from NL2SQLEvaluator.db_executor_nodes.sqlite_db_executor import SQLiteDBExecutor, _execute_single_query
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import register_node

logger = get_logger(__name__)

_TABLE_EXISTS = False


def create_cache_table(db_file) -> None:
    """Create the cache table if it does not exist."""
    create_table_sql = """
                       CREATE TABLE IF NOT EXISTS `cache_data`
                       (
                           `hash_key` TEXT PRIMARY KEY,
                           `db_id`    TEXT NOT NULL,
                           `query`    TEXT NOT NULL,
                           `result`   BLOB NOT NULL
                       ); \
                       """.strip()
    _execute_single_query(0, 0, db_file, create_table_sql, timeout=300, allow_write=True)
    _TABLE_EXISTS = True


@register_node()
class SqliteCache:
    @staticmethod
    def set_in_cache(db_file: str,
                     data_to_cache: list[DataToCache]) -> None:
        """Save a file with the given name and parameters."""
        # create the cache table if it does not exist:
        if not _TABLE_EXISTS:
            create_cache_table(db_file)
        insert_sql = """
                     INSERT OR IGNORE INTO `cache_data` (hash_key, db_id, query, result)
                     VALUES (:hash_key, :db_id, :query, :result);
                     """.strip()

        params = [
            data.model_dump(exclude={'result', 'dialect'}) | {'result': data.result.compress()}
            # This if is used to not store empty results which is used to detect if not found in cache
            if len(data.result) > 0 else data.model_dump(exclude={'result', 'dialect'}) | {
                'result': OutputTable(rows=[('empty',)]).compress()}
            for data in data_to_cache
        ]

        results = SQLiteDBExecutor.execute_queries(
            tasks=[ExecuteTask(
                db_files=db_file,
                queries=[insert_sql] * len(params),
                params=params,
            )],
            allow_write=True,
            timeout=3000,
        )[0][0]

        if isinstance(results, ExecutorError):
            logger.error(f'Impossible to set cache in the database. error: {results}')

    @staticmethod
    def get_from_cache(db_file: str, data_to_fetch: list[DataToFetch]) -> list[OutputTable | NotFoundInCacheError]:
        """Retrieve a file with the given name and parameters."""

        if not _TABLE_EXISTS:
            create_cache_table(db_file)

        select_sql = """
                     SELECT result
                     FROM `cache_data`
                     WHERE hash_key = :hash_key;
                     """.strip()
        params = [data.model_dump(include={'hash_key'}) for data in data_to_fetch]

        tasks = [ExecuteTask(
            db_files=db_file,
            queries=[select_sql] * len(params),
            params=params,
        )]
        results = SQLiteDBExecutor.execute_queries(
            tasks=tasks,
            allow_write=False,
            timeout=30,
        )[0]
        dec_results = []
        for data in results:
            decompressed = data.decompress()
            if len(decompressed) == 0:
                dec_results.append(NotFoundInCacheError())
            elif decompressed.rows == [('empty',)]:
                dec_results.append(OutputTable(rows=[]))
            else:
                dec_results.append(decompressed)
        return dec_results



import multiprocessing as mp
import sqlite3
from pathlib import Path
from typing import Any, TypeAlias

from func_timeout import func_timeout, FunctionTimedOut

from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import NotFoundInCacheError, OutputTable, SQLCacheProtocol, \
    DataToFetch
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError, ExecuteTask
from NL2SQLEvaluator.node_registry import register_node

ParamsType: TypeAlias = list[dict] | dict | None


@register_node()
class SQLiteDBExecutor:
    @staticmethod
    def execute_queries(
            tasks: list[ExecuteTask],
            cache_db: SQLCacheProtocol | None = None,
            cache_db_file: str | None = None,
            allow_write: bool = False,
            *args, **kwargs
    ) -> list[list[OutputTable | ExecutorError]]:

        timeout_s = kwargs.get("timeout", 10)
        tasks_mp = _build_task_for_mp(tasks, timeout_s, cache_db, cache_db_file, allow_write)
        total_tasks = len(tasks_mp)
        default_procs = min(max(1, mp.cpu_count()), max(1, total_tasks))
        num_cpus: int = kwargs.get("num_cpus", default_procs)
        if len(tasks_mp) == 1:
            return [[_execute_single_query(*tasks_mp[0])[2]]]
        # Run workers
        with mp.Pool(processes=num_cpus) as pool:
            flat_results: list[tuple[int, int, OutputTable | ExecutorError]] = pool.starmap(
                _execute_single_query,
                tasks_mp
            )
        # Reassemble into rectangular [jobs][idx] result
        results: list[list[OutputTable | ExecutorError]] = [[ExecutorError()] * len(task.queries) for task in tasks]
        for job_id, idx, value in flat_results:
            results[job_id][idx] = value

        return results


def _build_task_for_mp(tasks: list[ExecuteTask], timeout_s, cache_db, cache_db_file, allow_write):
    flattened_tasks = []
    for job_id, task in enumerate(tasks):
        for idx, (query, params, db_file) in enumerate(task):
            flattened_tasks.append(
                (job_id, idx, db_file, query, timeout_s, allow_write, params, cache_db, cache_db_file))
    return flattened_tasks


def _execute_single_query(
        job_id: int,
        idx: int,
        db_file: str,
        query: str,
        timeout: float,
        allow_write: bool = False,
        params: ParamsType = None,
        cache_db: SQLCacheProtocol | None = None,
        cache_db_file: str | None = None,
) -> tuple[int, int, OutputTable | ExecutorError]:
    """
    Execute a single SQL statement with a timeout, in an isolated connection.

    Behavior:
      - If a cache is provided and returns a hit, return the cached value.
      - If allow_write is False: PRAGMA query_only=ON, run the query, fetchall, then ROLLBACK.
      - If allow_write is True: run the query; if params is a list with len>1, use executemany;
        COMMIT on success and return cursor.rowcount.
      - On any exception: attempt ROLLBACK and return ExecutorError(e).
    """

    def _run() -> OutputTable | ExecutorError:
        # 1) Cache short-circuit
        cached = _maybe_get_cached(cache_db, cache_db_file, db_file, query)
        if cached is not None:
            return cached

        # 2) DB execution
        try:
            conn = sqlite3.connect(db_file)
            try:
                conn.execute("PRAGMA foreign_keys=ON;")
                if not allow_write:
                    conn.execute("PRAGMA query_only=ON;")

                conn.execute("BEGIN TRANSACTION;")
                cur = conn.cursor()

                if allow_write:
                    cur.execute(query, params or {})
                    conn.commit()
                    return OutputTable(rows=[([cur.rowcount])])

                else:
                    # Reads: always single execute + fetchall + rollback
                    cur.execute(query, params or {})
                    rows = cur.fetchall()
                    conn.rollback()
                    return OutputTable(rows=rows)

            except Exception as e:
                # Try rollback, ignore rollback failures
                try:
                    conn.rollback()
                except Exception:
                    pass
                return ExecutorError(e)

            finally:
                conn.close()

        except Exception as e:
            # Connection-level failures
            return ExecutorError(e)

    try:
        result = func_timeout(timeout, _run)
    except FunctionTimedOut:
        result = ExecutorError(f"Query Timeout with {timeout} seconds")

    return job_id, idx, result


# -------------------------
# Small, focused helpers
# -------------------------

def _db_id_from_path(db_file: str) -> str:
    """Derive a stable DB id from a file path (matches previous basename-without-ext logic)."""
    return Path(db_file).stem


def _maybe_get_cached(
        cache_db: SQLCacheProtocol | None,
        cache_db_file: str | None,
        db_file: str,
        query: str,
) -> Any | None:
    """Return cached value if available; None on miss or when cache is not configured."""
    if cache_db is None or cache_db_file is None:
        return None
    db_id = _db_id_from_path(db_file)
    cached = cache_db.get_from_cache(cache_db_file, data_to_fetch=[DataToFetch(db_id=db_id, query=query)])[0]
    if cached and not isinstance(cached, NotFoundInCacheError):
        return cached
    return None



from collections import Counter

from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable


def sort_with_different_types(arr: tuple) -> tuple:
    return tuple(sorted(arr, key=sort_key))


def sort_key(x):
    raw = x
    if raw is None:
        return 0, ''
    elif isinstance(raw, (int, float)):
        return 1, float(raw)
    else:
        return 2, str(raw)


def get_majority_voting_values(values: list[OutputTable], count_cardinality_in_row: bool) -> OutputTable | None:
    counter = Counter()
    for pred_n in values:
        if pred_n is None or isinstance(pred_n, Exception):
            continue

        if count_cardinality_in_row:
            key = tuple([sort_with_different_types(row) for row in pred_n])

        else:
            key = frozenset(pred_n)

        counter[key] += 1

    if not counter:
        # case where all the predictions are None
        return None

    majority_vote, _ = counter.most_common(1)[0]
    return majority_vote



import pytest

from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable
from NL2SQLEvaluator.evaluator_nodes import BirdEXEvaluator
from NL2SQLEvaluator.evaluator_nodes.evaluator_protocol import EvaluateTask


@pytest.fixture
def executor() -> BirdEXEvaluator:
    return BirdEXEvaluator()


class TestBirdEX:
    def _internal_run(self, tar, pred, executor):
        task = EvaluateTask(
            predictions=[OutputTable(rows=pred[0][0])],
            target=[OutputTable(rows=tar[0][0])]
        )

        result = executor.execute_metric(
            tasks=[task]
        )
        return result

    def test_equal(self, executor):
        multiple_tasks_tar = [('a', 'b'), ('c', 'd')]
        multiple_tasks_preds = [('a', 'b'), ('c', 'd')]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0

    def test_different(self, executor):
        multiple_tasks_tar = [('a', 'b'), ('c', 'd')]
        multiple_tasks_preds = [('c', 'd')]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 0.0

    def test_equal_but_different_projection(self, executor):
        multiple_tasks_tar = [('a', 'b'), ('c', 'd')]
        multiple_tasks_preds = [('b', 'a'), ('d', 'c')]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 0.0

    def test_equal_but_different_tuple_order(self, executor):
        multiple_tasks_tar = [('a', 'b'), ('c', 'd')]
        multiple_tasks_preds = [('c', 'd'), ('a', 'b')]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0

    def test_null_values(self, executor):
        multiple_tasks_tar = [('a', None), ('c', 'd')]
        multiple_tasks_preds = [('c', 'd'), ('a', None)]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0

    def test_null_math_none(self, executor):
        import math
        multiple_tasks_tar = [('a', math.nan), ('c', 'd')]
        multiple_tasks_preds = [('c', 'd'), ('a', math.nan)]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0

    def test_mixed_types(self, executor):
        multiple_tasks_tar = [('a', 1), ('c', 'd')]
        multiple_tasks_preds = [('c', 'd'), ('a', 1.0000000001)]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 0.0

    def test_empty_lists(self, executor):
        multiple_tasks_tar = []
        multiple_tasks_preds = []
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0

    def test_empty_vs_non_empty(self, executor):
        multiple_tasks_tar = []
        multiple_tasks_preds = [('a',)]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 0.0

    def test_nan_and_string_mismatch(self, executor):
        import math
        multiple_tasks_tar = [(math.nan,)]
        multiple_tasks_preds = [('NaN',)]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 0.0

    def test_bird_ex_no_distinct(self, executor):
        multiple_tasks_tar = [('a', 'b'), ('a', 'b')]
        multiple_tasks_preds = [('a', 'b')]
        result = self._internal_run([[multiple_tasks_tar]], [[multiple_tasks_preds]], executor)
        assert result[0] == 1.0



from collections import Counter
from enum import Enum

from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable
from NL2SQLEvaluator.evaluator_nodes.evaluator_protocol import EvaluateTask
from NL2SQLEvaluator.evaluator_nodes.utils import sort_with_different_types, get_majority_voting_values
from NL2SQLEvaluator.node_registry import register_node


class QatchMetric(Enum):
    """Enum for different metrics used in evaluation."""
    CELL_PRECISION = "cell_precision"
    CELL_RECALL = "cell_recall"
    TUPLE_CARDINALITY = "tuple_cardinality"
    TUPLE_CONSTRAINT = "tuple_constraint"
    TUPLE_ORDER = "tuple_order"
    CP_CR_TC = "cp_cr_tc"
    F1_SCORE = "f1_score"

    # Optional: accept case-insensitive and aliases like "val" / "validation"
    _ALIASES = {
        'f1_score': 'F1_SCORE',
        'cell_precision': 'CELL_PRECISION',
        'cell_recall': 'CELL_RECALL',
        'tuple_cardinality': 'TUPLE_CARDINALITY',
        'tuple_constraint': 'TUPLE_CONSTRAINT',
        'tuple_order': 'TUPLE_ORDER',
        'cp_cr_tc': 'CP_CR_TC',
    }

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            v = value.strip().lower()
            v = cls._ALIASES.get(v, v)
            for m in cls:
                if m.value == v or m.name == v:
                    return m
        return None


@register_node()
class QATCHEvaluator:
    def execute_metric(
            self,
            tasks: list[EvaluateTask],
            metric: QatchMetric | str,
            *args,
            **kwargs
    ) -> list[float]:
        results = []
        metric = QatchMetric(metric)
        for task in tasks:
            tar = task.target[0]
            pred = task.predictions

            tar = OutputTable(rows=[tuple(sort_with_different_types(row)) for row in tar])
            majority_vote = get_majority_voting_values(pred, count_cardinality_in_row=True)

            if majority_vote is None:
                results.append(0.0)
            elif len(majority_vote) == len(tar) == 0:
                results.append(1.0)
            elif len(majority_vote) != 0 and len(tar) == 0 or len(majority_vote) == 0 and len(tar) != 0:
                results.append(0.0)

            elif metric == QatchMetric.TUPLE_CARDINALITY:
                results.append(self._tuple_cardinality(majority_vote, tar))
            elif metric == QatchMetric.TUPLE_CONSTRAINT:
                results.append(self._tuple_constraint(majority_vote, tar))
            elif metric == QatchMetric.CELL_PRECISION:
                results.append(self._cell_precision(majority_vote, tar))
            elif metric == QatchMetric.CELL_RECALL:
                results.append(self._cell_recall(majority_vote, tar))
            elif metric == QatchMetric.TUPLE_ORDER:
                results.append(self._tuple_order(majority_vote, tar))
            elif metric == QatchMetric.CP_CR_TC:
                results.append(self._cp_cr_tc(majority_vote, tar))
            elif metric == QatchMetric.F1_SCORE:
                results.append(self._f1_score(majority_vote, tar))
            else:
                raise NotImplementedError(f"The metric {metric} is not implemented in QATCHEvaluator.")

        return results

    def _tuple_cardinality(self, task_pred: OutputTable, task_tar: OutputTable) -> float:

        if len(task_pred) >= len(task_tar):
            # in case we have more elements in the prediction than in the target
            return round(len(task_tar) / len(task_pred), 3)

            # in case we have more elements in the target than in the prediction
        return round(len(task_pred) / len(task_tar), 3)

    def _tuple_constraint(self, task_pred: OutputTable, task_tar: OutputTable):
        target_len = len(task_tar)
        prediction_len = len(task_pred)
        if target_len == prediction_len == 0:
            return 1.0
        if prediction_len != 0 and target_len == 0 or prediction_len == 0 and target_len != 0:
            return 0.0

        # When comparing tuples, the projection orders do not matter (Name, Surname) = (Surname, Name)

        count_targ_dict = Counter(task_tar)
        count_pred_dict = Counter(task_pred)

        cardinality = [count_pred_dict[key] == count for key, count in count_targ_dict.items()]

        return round(sum(cardinality) / len(cardinality), 3)

    def _cell_precision(self, task_pred: OutputTable, task_tar: OutputTable) -> float:
        from itertools import chain

        prediction = set(chain.from_iterable(task_pred))
        target = set(chain.from_iterable(task_tar))
        intersected_cells = target.intersection(prediction)

        return round(len(intersected_cells) / len(prediction), 3) if len(prediction) > 0 else 0.0

    def _cell_recall(self, task_pred: OutputTable, task_tar: OutputTable) -> float:
        from itertools import chain

        prediction = set(chain.from_iterable(task_pred))
        target = set(chain.from_iterable(task_tar))
        intersected_cells = target.intersection(prediction)

        return round(len(intersected_cells) / len(target), 3) if len(target) > 0 else 0.0

    def _tuple_order(self, task_pred: OutputTable, task_tar: OutputTable) -> float:
        import numpy as np
        def normalize(data: float):
            data = [-1, data, 1]
            data = (data - np.min(data)) / (np.max(data) - np.min(data))
            return data[1]

        new_pred = []
        [new_pred.append(row) for row in task_pred
         if row in task_tar and row not in new_pred]

        # same for target
        new_target = []
        [new_target.append(row) for row in task_tar
         if row in task_pred and row not in new_target]

        if len(new_target) == 0:
            # case when prediction does not have any element in target
            rho = 0.0

        else:
            target_ranks = [i for i in range(len(new_target))]
            pred_ranks = [new_target.index(row) for row in new_pred]

            diff_rank_squared = [(tar - pred) ** 2
                                 for tar, pred in zip(target_ranks, pred_ranks)]

            sum_diff_rank_squared = sum(diff_rank_squared)

            n = len(new_target) if len(new_target) > 1 else 2
            rho = 1 - 6 * sum_diff_rank_squared / (n * (n ** 2 - 1))

        return float(normalize(round(rho, 3)))

    def _f1_score(self, task_pred: OutputTable, task_tar: OutputTable) -> float:
        cell_precision = self._cell_precision(task_pred, task_tar)
        cell_recall = self._cell_recall(task_pred, task_tar)
        if cell_precision + cell_recall == 0:
            return 0.0

        f1_score = 2 * (cell_precision * cell_recall) / (cell_precision + cell_recall)
        return float(f1_score)

    def _cp_cr_tc(self, task_pred: OutputTable, task_tar: OutputTable) -> float:
        cell_precision = self._cell_precision(task_pred, task_tar)
        cell_recall = self._cell_recall(task_pred, task_tar)
        tuple_cardinality = self._tuple_cardinality(task_pred, task_tar)

        combined_score = (cell_precision + cell_recall + tuple_cardinality) / 3
        return round(combined_score, 3)

    @staticmethod
    def _validate_inputs(multiple_tasks_preds, multiple_tasks_tar) -> None:
        if not (len(multiple_tasks_preds) == len(multiple_tasks_tar)):
            raise ValueError(
                "Lengths must match: target_queries, llm_predictions, db_files."
            )
        for ref in multiple_tasks_tar:
            if not len(ref) == 1:
                raise ValueError(
                    "For Execution Accuracy the target must have len 1. Otherwise use Ambiguous Execution Accuracy."
                )
